package ru.dbotthepony.kstarbound

import com.google.common.collect.Interner
import com.google.common.collect.Interners
import com.google.gson.*
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.api.IStarboundFile
import ru.dbotthepony.kstarbound.api.NonExistingFile
import ru.dbotthepony.kstarbound.api.PhysicalFile
import ru.dbotthepony.kstarbound.api.explore
import ru.dbotthepony.kstarbound.defs.*
import ru.dbotthepony.kstarbound.defs.image.ImageReference
import ru.dbotthepony.kstarbound.defs.image.SpriteReference
import ru.dbotthepony.kstarbound.defs.item.BackArmorItemPrototype
import ru.dbotthepony.kstarbound.defs.item.ChestArmorItemPrototype
import ru.dbotthepony.kstarbound.defs.item.CurrencyItemPrototype
import ru.dbotthepony.kstarbound.defs.item.FlashlightPrototype
import ru.dbotthepony.kstarbound.defs.item.HarvestingToolPrototype
import ru.dbotthepony.kstarbound.defs.item.HeadArmorItemPrototype
import ru.dbotthepony.kstarbound.defs.item.IArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.IItemDefinition
import ru.dbotthepony.kstarbound.defs.item.ItemPrototype
import ru.dbotthepony.kstarbound.defs.item.LegsArmorItemPrototype
import ru.dbotthepony.kstarbound.defs.item.LiquidItemPrototype
import ru.dbotthepony.kstarbound.defs.item.MaterialItemPrototype
import ru.dbotthepony.kstarbound.defs.liquid.LiquidDefinition
import ru.dbotthepony.kstarbound.defs.parallax.ParallaxPrototype
import ru.dbotthepony.kstarbound.defs.particle.ParticleDefinition
import ru.dbotthepony.kstarbound.defs.player.PlayerDefinition
import ru.dbotthepony.kstarbound.defs.projectile.*
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.io.*
import ru.dbotthepony.kstarbound.io.json.EitherTypeAdapter
import ru.dbotthepony.kstarbound.io.json.builder.EnumAdapter
import ru.dbotthepony.kstarbound.io.json.builder.BuilderAdapter
import ru.dbotthepony.kstarbound.io.json.builder.FactoryAdapter
import ru.dbotthepony.kstarbound.io.json.builder.JsonImplementationTypeFactory
import ru.dbotthepony.kstarbound.io.json.factory.ArrayListAdapterFactory
import ru.dbotthepony.kstarbound.io.json.factory.ImmutableCollectionAdapterFactory
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.WriteOnce
import java.io.*
import java.text.DateFormat
import java.util.*
import java.util.function.BiConsumer
import java.util.function.BinaryOperator
import java.util.function.Function
import java.util.function.Supplier
import java.util.stream.Collector
import kotlin.collections.ArrayList

const val METRES_IN_STARBOUND_UNIT = 0.5
const val METRES_IN_STARBOUND_UNITf = 0.5f

const val PIXELS_IN_STARBOUND_UNIT = 8.0
const val PIXELS_IN_STARBOUND_UNITf = 8.0f

// class TileDefLoadingException(message: String, cause: Throwable? = null) : IllegalStateException(message, cause)
// class ProjectileDefLoadingException(message: String, cause: Throwable? = null) : IllegalStateException(message, cause)

fun String.sbIntern(): String = Starbound.STRING_INTERNER.intern(this)

object Starbound {
	private val LOGGER = LogManager.getLogger()

	val STRING_INTERNER: Interner<String> = Interners.newWeakInterner()
	val pathStack = AssetPathStack(STRING_INTERNER)

	fun assetFolder(input: String): String {
		return pathStack.remap(input)
	}

	private val tiles = Object2ObjectOpenHashMap<String, TileDefinition>()
	private val tilesByMaterialID = Int2ObjectOpenHashMap<TileDefinition>()

	private val tileModifiers = Object2ObjectOpenHashMap<String, MaterialModifier>()
	private val tileModifiersByID = Int2ObjectOpenHashMap<MaterialModifier>()

	private val liquid = Object2ObjectOpenHashMap<String, LiquidDefinition>()
	private val liquidByID = Int2ObjectOpenHashMap<LiquidDefinition>()

	private val projectiles = Object2ObjectOpenHashMap<String, ConfiguredProjectile>()
	private val parallax = Object2ObjectOpenHashMap<String, ParallaxPrototype>()
	private val functions = Object2ObjectOpenHashMap<String, JsonFunction>()
	private val species = Object2ObjectOpenHashMap<String, Species>()
	private val _statusEffects = Object2ObjectOpenHashMap<String, StatusEffectDefinition>()
	private val _particles = Object2ObjectOpenHashMap<String, ParticleDefinition>()

	val particles: Map<String, ParticleDefinition> = Collections.unmodifiableMap(_particles)
	val statusEffects: Map<String, StatusEffectDefinition> = Collections.unmodifiableMap(_statusEffects)

	private val items = Object2ObjectOpenHashMap<String, IItemDefinition>()

	val LIQUID: Map<String, LiquidDefinition> = Collections.unmodifiableMap(liquid)
	val LIQUID_BY_ID: Map<Int, LiquidDefinition> = Collections.unmodifiableMap(liquidByID)
	val TILE_MODIFIER: Map<String, MaterialModifier> = Collections.unmodifiableMap(tileModifiers)
	val TILE_MODIFIER_BY_ID: Map<Int, MaterialModifier> = Collections.unmodifiableMap(tileModifiersByID)
	val TILE: Map<String, TileDefinition> = Collections.unmodifiableMap(tiles)
	val TILE_BY_ID: Map<Int, TileDefinition> = Collections.unmodifiableMap(tilesByMaterialID)
	val PROJECTILE: Map<String, ConfiguredProjectile> = Collections.unmodifiableMap(projectiles)
	val PARALLAX: Map<String, ParallaxPrototype> = Collections.unmodifiableMap(parallax)
	val FUNCTION: Map<String, JsonFunction> = Collections.unmodifiableMap(functions)
	val ITEM: Map<String, IItemDefinition> = Collections.unmodifiableMap(items)

	val STRING_ADAPTER: TypeAdapter<String> = object : TypeAdapter<String>() {
		override fun write(out: JsonWriter, value: String?) {
			if (value == null)
				out.nullValue()
			else
				out.value(value)
		}

		override fun read(`in`: JsonReader): String? {
			return STRING_INTERNER.intern(TypeAdapters.STRING.read(`in`) ?: return null)
		}
	}

	val GSON: Gson = GsonBuilder()
		.enableComplexMapKeySerialization()
		.serializeNulls()
		.setDateFormat(DateFormat.LONG)
		.setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
		.setPrettyPrinting()

		// чтоб строки всегда intern'ились
		.registerTypeAdapter(STRING_ADAPTER)

		// Обработчик @JsonImplementation
		.registerTypeAdapterFactory(JsonImplementationTypeFactory)

		// ImmutableList, ImmutableSet, ImmutableMap
		.registerTypeAdapterFactory(ImmutableCollectionAdapterFactory)

		// ArrayList
		.registerTypeAdapterFactory(ArrayListAdapterFactory)

		// все enum'ы без особых настроек
		.registerTypeAdapterFactory(EnumAdapter.Companion)

		// автоматическое создание BuilderAdapter по @аннотациям
		.registerTypeAdapterFactory(BuilderAdapter.Companion)

		// автоматическое создание FactoryAdapter по @аннотациям
		.registerTypeAdapterFactory(FactoryAdapter.Companion)

		.registerTypeAdapterFactory(EitherTypeAdapter)

		.also(::addStarboundJsonAdapters)

		.registerTypeAdapterFactory(IItemDefinition.InventoryIcon.Factory(pathStack))
		.registerTypeAdapter(SpriteReference.Adapter(pathStack))
		.registerTypeAdapterFactory(IArmorItemDefinition.ArmorFrames.Factory(pathStack))
		.registerTypeAdapter(ImageReference.Adapter(pathStack))

		.registerTypeAdapterFactory(AssetReferenceFactory(pathStack) {
			val file = locate(it)

			if (file.exists && file.isFile)
				file.reader()
			else
				null
		})

		.registerTypeAdapterFactory(RegistryReferenceFactory()
										.add(tiles::get)
										.add(tileModifiers::get)
										.add(liquid::get)
										.add(projectiles::get)
										.add(parallax::get)
										.add(functions::get)
										.add(items::get)
										.add(species::get)
										.add(_statusEffects::get)
										.add(_particles::get)
		)

		.create()

	@Suppress("unchecked_cast")
	fun <T> getTypeAdapter(type: Class<T>): TypeAdapter<T> {
		return when (type) {
			Float::class.java -> TypeAdapters.FLOAT as TypeAdapter<T>
			Double::class.java -> TypeAdapters.DOUBLE as TypeAdapter<T>
			String::class.java -> STRING_ADAPTER as TypeAdapter<T>
			Int::class.java -> TypeAdapters.INTEGER as TypeAdapter<T>
			Long::class.java -> TypeAdapters.LONG as TypeAdapter<T>
			Boolean::class.java -> TypeAdapters.BOOLEAN as TypeAdapter<T>
			else -> GSON.getAdapter(type) as TypeAdapter<T>
		}
	}

	var initializing = false
		private set
	var initialized = false
		private set

	@Volatile
	var terminateLoading = false

	private val archivePaths = ArrayList<File>()
	private val fileSystems = ArrayList<IStarboundFile>()

	fun addFilePath(path: File) {
		fileSystems.add(PhysicalFile(path))
	}

	fun addPak(pak: StarboundPak) {
		fileSystems.add(pak.root)
	}

	fun exists(path: String): Boolean {
		@Suppress("name_shadowing")
		var path = path

		if (path[0] == '/') {
			path = path.substring(1)
		}

		for (fs in fileSystems) {
			if (fs.locate(path).exists) {
				return true
			}
		}

		return false
	}

	fun locate(path: String): IStarboundFile {
		@Suppress("name_shadowing")
		var path = path

		if (path[0] == '/') {
			path = path.substring(1)
		}

		for (fs in fileSystems) {
			val file = fs.locate(path)

			if (file.exists) {
				return file
			}
		}

		return NonExistingFile(path.split("/").last(), fullPath = path)
	}

	fun locate(vararg path: String): IStarboundFile {
		for (p in path) {
			val get = locate(p)

			if (get.exists) {
				return get
			}
		}

		return NonExistingFile(path[0].split("/").last(), fullPath = path[0])
	}

	/**
	 * Добавляет pak к чтению при initializeGame
	 */
	fun addPakPath(pak: File) {
		archivePaths.add(pak)
	}

	fun loadJson(path: String): JsonElement {
		return JsonParser.parseReader(locate(path).reader())
	}

	fun readDirect(path: String) = locate(path).readDirect()

	fun getTileDefinition(name: String) = tiles[name]
	private val initCallbacks = ArrayList<() -> Unit>()

	var playerDefinition: PlayerDefinition by WriteOnce()
		private set

	private fun loadStage(
		callback: (Boolean, Boolean, String) -> Unit,
		loader: ((String) -> Unit) -> Unit,
		name: String,
	) {
		if (terminateLoading)
			return

		val time = System.currentTimeMillis()
		callback(false, false, "Loading $name...")
		LOGGER.info("Loading $name...")

		loader {
			if (terminateLoading) {
				throw InterruptedException("Game is terminating")
			}

			callback(false, true, it)
		}

		callback(false, true, "Loaded $name in ${System.currentTimeMillis() - time}ms")
		LOGGER.info("Loaded $name in ${System.currentTimeMillis() - time}ms")
	}

	private fun <T, K : Any> loadStage(
		callback: (Boolean, Boolean, String) -> Unit,
		clazz: Class<T>,
		getKey: (T) -> K,
		put: (K, T) -> T?,
		files: List<IStarboundFile>,
		name: String,
	) {
		loadStage(callback, loader = {
			for (listedFile in files) {
				try {
					it("Loading $listedFile")
					val def = pathStack(listedFile.computeDirectory()) { GSON.fromJson(listedFile.reader(), clazz) }
					val key = getKey.invoke(def)
					check(put.invoke(key, def) == null) { "Already has $name with name $key loaded!" }
				} catch (err: Throwable) {
					LOGGER.error("Loading $name definition file $listedFile", err)
				}

				if (terminateLoading) {
					break
				}
			}
		}, name)
	}

	private fun <T, K0 : Any, K1 : Any> loadStage(
		callback: (Boolean, Boolean, String) -> Unit,
		clazz: Class<T>,
		getKey0: (T) -> K0,
		put0: (K0, T) -> T?,
		getKey1: (T) -> K1,
		put1: (K1, T) -> T?,
		files: List<IStarboundFile>,
		name: String,
	) {
		loadStage(callback, loader = {
			for (listedFile in files) {
				try {
					it("Loading $listedFile")
					val def = pathStack(listedFile.computeDirectory()) { GSON.fromJson(listedFile.reader(), clazz) }
					val key0 = getKey0.invoke(def)
					val key1 = getKey1.invoke(def)
					check(put0.invoke(key0, def) == null) { "Already has $name with name $key0 loaded!" }
					check(put1.invoke(key1, def) == null) { "Already has $name with ID $key1 loaded!" }
				} catch (err: Throwable) {
					LOGGER.error("Loading $name definition file $listedFile", err)
				}

				if (terminateLoading) {
					break
				}
			}
		}, name)
	}

	private fun doInitialize(callback: (finished: Boolean, replaceStatus: Boolean, status: String) -> Unit) {
		var time = System.currentTimeMillis()

		if (archivePaths.isNotEmpty()) {
			callback(false, false, "Searching for pak archives...".also(LOGGER::info))

			for (path in archivePaths) {
				callback(false, false, "Reading index of ${path.name}...".also(LOGGER::info))

				addPak(StarboundPak(path) { _, status ->
					callback(false, true, "${path.name}: $status")
				})
			}
		}

		callback(false, false, "Finished reading pak archives in ${System.currentTimeMillis() - time}ms".also(LOGGER::info))
		time = System.currentTimeMillis()
		callback(false, false, "Building file index...".also(LOGGER::info))

		val ext2files = fileSystems.parallelStream()
			.flatMap { it.explore() }
			.filter { it.isFile }
			.collect(object :
					Collector<IStarboundFile, Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>,
					Map<String, List<IStarboundFile>>>,
					Supplier<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>>,
					BiConsumer<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>, IStarboundFile>,
					BinaryOperator<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>>
			{
				override fun accept(t: Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>, u: IStarboundFile) {
					t.computeIfAbsent(u.name.substringAfterLast('.'), Object2ObjectFunction { ArrayList() }).add(u)
				}

				override fun get(): Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>> {
					return Object2ObjectOpenHashMap()
				}

				override fun supplier(): Supplier<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>> {
					return this
				}

				override fun accumulator(): BiConsumer<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>, IStarboundFile> {
					return this
				}

				override fun apply(
					t: Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>,
					u: Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>
				): Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>> {
					t.putAll(u)
					return t
				}

				override fun combiner(): BinaryOperator<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>> {
					return this
				}

				override fun finisher(): Function<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>, Map<String, List<IStarboundFile>>> {
					return Function { it }
				}

				override fun characteristics(): Set<Collector.Characteristics> {
					return setOf(Collector.Characteristics.IDENTITY_FINISH, Collector.Characteristics.UNORDERED)
				}
			})

		callback(false, false, "Finished building file index in ${System.currentTimeMillis() - time}ms".also(LOGGER::info))

		loadStage(callback, this::loadFunctions, "functions")
		loadStage(callback, this::loadProjectiles, "projectiles")
		loadStage(callback, this::loadParallax, "parallax definitions")
		loadStage(callback, this::loadItemDefinitions, "item definitions")

		loadStage(callback, TileDefinition::class.java, TileDefinition::materialName, tiles::put, TileDefinition::materialId, tilesByMaterialID::put, ext2files["material"] ?: listOf(), "materials")
		loadStage(callback, MaterialModifier::class.java, MaterialModifier::modName, tileModifiers::put, MaterialModifier::modId, tileModifiersByID::put, ext2files["matmod"] ?: listOf(), "material modifier definitions")
		loadStage(callback, LiquidDefinition::class.java, LiquidDefinition::name, liquid::put, LiquidDefinition::liquidId, liquidByID::put, ext2files["liquid"] ?: listOf(), "liquid definitions")
		loadStage(callback, StatusEffectDefinition::class.java, StatusEffectDefinition::name, _statusEffects::put, ext2files["statuseffect"] ?: listOf(), "status effects")
		loadStage(callback, Species::class.java, Species::kind, species::put, ext2files["species"] ?: listOf(), "species")
		loadStage(callback, ParticleDefinition::class.java, ParticleDefinition::kind, _particles::put, ext2files["particle"] ?: listOf(), "particles")

		pathStack.block("/") {
			playerDefinition = GSON.fromJson(locate("/player.config").reader(), PlayerDefinition::class.java)
		}

		initializing = false
		initialized = true
		callback(true, false, "Finished loading in ${System.currentTimeMillis() - time}ms")
	}

	fun initializeGame(callback: (finished: Boolean, replaceStatus: Boolean, status: String) -> Unit) {
		if (initializing) {
			throw IllegalStateException("Already initializing!")
		}

		if (initialized) {
			throw IllegalStateException("Already initialized!")
		}

		initializing = true
		Thread({ doInitialize(callback) }, "Asset Loader").start()
	}

	fun onInitialize(callback: () -> Unit) {
		if (initialized) {
			callback()
		} else {
			initCallbacks.add(callback)
		}
	}

	fun pollCallbacks() {
		if (initialized && initCallbacks.isNotEmpty()) {
			for (callback in initCallbacks) {
				callback()
			}

			initCallbacks.clear()
		}
	}

	private fun loadProjectiles(callback: (String) -> Unit) {
		for (fs in fileSystems) {
			for (listedFile in fs.explore().filter { it.isFile }.filter { it.name.endsWith(".projectile") }) {
				try {
					callback("Loading $listedFile")

					val def = pathStack(listedFile.computeDirectory()) { GSON.fromJson(listedFile.reader(), ConfigurableProjectile::class.java).assemble(it) }
					check(projectiles[def.projectileName] == null) { "Already has projectile with ID ${def.projectileName} loaded!" }
					projectiles[def.projectileName] = def
				} catch(err: Throwable) {
					//throw ProjectileDefLoadingException("Loading projectile file $listedFile", err)
					LOGGER.error("Loading projectile file $listedFile", err)
				}

				if (terminateLoading) {
					return
				}
			}
		}
	}

	private fun loadFunctions(callback: (String) -> Unit) {
		for (fs in fileSystems) {
			for (listedFile in fs.explore().filter { it.isFile }.filter { it.name.endsWith(".functions") }) {
				try {
					callback("Loading $listedFile")

					val readObject = JsonParser.parseReader(listedFile.reader()) as JsonObject

					for (key in readObject.keySet()) {
						val def = pathStack(listedFile.computeDirectory()) { GSON.fromJson(readObject[key], JsonFunction::class.java) }
						functions[key] = def
					}
				} catch(err: Throwable) {
					LOGGER.error("Loading function file $listedFile", err)
				}

				if (terminateLoading) {
					return
				}
			}
		}
	}

	private fun loadParallax(callback: (String) -> Unit) {
		for (fs in fileSystems) {
			for (listedFile in fs.explore().filter { it.isFile }.filter { it.name.endsWith(".parallax") }) {
				try {
					callback("Loading $listedFile")
					val def = pathStack(listedFile.computeDirectory()) { GSON.fromJson(listedFile.reader(), ParallaxPrototype::class.java) }
					parallax[listedFile.name.substringBefore('.')] = def
				} catch(err: Throwable) {
					LOGGER.error("Loading parallax file $listedFile", err)
				}

				if (terminateLoading) {
					return
				}
			}
		}
	}

	private fun loadItemDefinitions(callback: (String) -> Unit) {
		val files = linkedMapOf(
			".item" to ItemPrototype::class.java,
			".currency" to CurrencyItemPrototype::class.java,
			".liqitem" to LiquidItemPrototype::class.java,
			".matitem" to MaterialItemPrototype::class.java,
			".flashlight" to FlashlightPrototype::class.java,
			".harvestingtool" to HarvestingToolPrototype::class.java,
			".head" to HeadArmorItemPrototype::class.java,
			".chest" to ChestArmorItemPrototype::class.java,
			".legs" to LegsArmorItemPrototype::class.java,
			".back" to BackArmorItemPrototype::class.java,
		)

		for (fs in fileSystems) {
			for (listedFile in fs.explore().filter { it.isFile }.filter { f -> files.keys.any { f.name.endsWith(it) } }) {
				try {
					callback("Loading $listedFile")
					val def: ItemPrototype = pathStack(listedFile.computeDirectory()) { GSON.fromJson(listedFile.reader(), files.entries.first { listedFile.name.endsWith(it.key) }.value) }
					check(items.put(def.itemName, def.assemble()) == null) { "Already has item with name ${def.itemName} loaded!" }
				} catch (err: Throwable) {
					LOGGER.error("Loading item definition file $listedFile", err)
				}

				if (terminateLoading) {
					return
				}
			}
		}
	}
}
